﻿using System;
using System.Linq;
using Sdl.Core.Globalization;
using Sdl.Core.Settings;
using Sdl.FileTypeSupport.Framework.BilingualApi;
using Sdl.FileTypeSupport.Framework.Core.Utilities.NativeApi;
using Sdl.FileTypeSupport.Framework.IntegrationApi;
using Sdl.FileTypeSupport.Framework.NativeApi;
using System.Xml;
using System.Xml.Linq;
using System.Xml.XPath;
using Xliff20FileType.Parser;
using System.Collections.Generic;
using Xliff20FileType.Common;
using System.IO;

namespace Xliff20FileType
{
    public class XliffParser : AbstractBilingualFileTypeComponent, IBilingualParser, INativeContentCycleAware, ISettingsAware
    {
        #region Private fields
        
        private IFileProperties _fileProperties;
        private XmlReader _reader;
        private BcmGeneratorRepository _bcmGeneratorRepository;
        private Stack<ContextData> _contextStack = new Stack<ContextData>();
        
        #endregion Private fields

        #region Main
        public void StartOfInput()
        {
            // create reader and jump to xliff element
            _reader = XmlReader.Create(_fileProperties.FileConversionProperties.InputFilePath);
            _reader.MoveToContent();

            // xliff properties
            string sourceLanguage = _reader.GetAttribute("srcLang");
            DocumentProperties.SourceLanguage = new Language(sourceLanguage);
            string targetLanguage = _reader.GetAttribute("tgtLang");
            if (!String.IsNullOrEmpty(targetLanguage))
            {
                DocumentProperties.TargetLanguage = new Language(targetLanguage);
            }

            // create bcm generators
            _bcmGeneratorRepository = new BcmGeneratorRepository();
            _bcmGeneratorRepository.ItemFactory = ItemFactory;
            _bcmGeneratorRepository.PropertiesFactory = PropertiesFactory;
        }

        public bool ParseNext()
        {
            while (_reader.Read())
            {
                if (_reader.LocalName == "file")
                {
                    if (_reader.NodeType == XmlNodeType.Element)
                    {
                        IContextInfo contextInfo = PropertiesFactory.CreateContextInfo("file");
                        contextInfo.DisplayCode = "FL";
                        contextInfo.DisplayName = Path.GetFileName(_reader.GetAttribute("original"));
                        contextInfo.Description = _reader.GetAttribute("original");
                        string fsValue = _reader.GetAttribute("fs", XliffConstants.Xliff20FormatStyleNamespace);
                        contextInfo.SetMetaData("fs", String.IsNullOrEmpty(fsValue) ? "div" : fsValue);
                        ContextData fileContext = new ContextData
                        {
                            Context = contextInfo,
                            StructureInfo = PropertiesFactory.CreateStructureInfo(contextInfo, true, null)                            
                        };
                        _contextStack.Push(fileContext);
                    }
                    else if (_reader.NodeType == XmlNodeType.EndElement)
                    {
                        _contextStack.Pop();
                    }
                }
                if (_reader.LocalName == "group")
                {
                    if (_reader.NodeType == XmlNodeType.Element)
                    {
                        StartGroup();
                    }
                    else if (_reader.NodeType == XmlNodeType.EndElement)
                    {
                        EndGroup();
                    }
                }
                else if (_reader.LocalName == "unit")
                {
                    XElement unitElement = (XElement)XElement.ReadFrom(_reader);
                    ProcessUnit(unitElement);
                   
                    return true;
                }
            }

            return false;
        }

        #endregion Main

        #region <unit> element

        private void ProcessUnit(XElement unitElement)
        {
            // translate=[yes|no]
            string translateAttribute = (string)unitElement.Attribute("translate") ?? "yes";
            LockTypeFlags lockType =   translateAttribute == "no"  ?  LockTypeFlags.Structure : LockTypeFlags.Unlocked;
            
            // create paragraph unit
            IParagraphUnit paragraphUnit = ItemFactory.CreateParagraphUnit(lockType);
            
            // id
            paragraphUnit.Properties.ParagraphUnitId = new ParagraphUnitId(unitElement.Attribute("id").Value);


            UnitInfo unitInfo = new UnitInfo
            {
                UnitElement = unitElement,
                OriginalData = ParseOriginalData(unitElement.Element(XliffConstants.originalData))
            };

            // segments and ignorables
            foreach (XElement child in unitElement.Elements())
            {
                switch (child.Name.LocalName)
                {
                    case "segment":
                        CreateSegmentPair(unitInfo, child, paragraphUnit);
                        break;
                    case "ignorable":
                        CreateContentOutsideSegments(unitInfo, child, paragraphUnit);
                        break;
                    case "notes":
                        paragraphUnit.Properties.Comments = CreateUnitComments(child);
                        break;
                    default:
                        // ignore
                        break;
                }
            }

            // context
            SetContextProperties(paragraphUnit, unitElement);
           
            // send to output
            Output.ProcessParagraphUnit(paragraphUnit);
        }

        private void CreateSegmentPair(UnitInfo unitInfo, XElement segmentElement, IParagraphUnit paragraphUnit)
        {
            XElement sourceElement = segmentElement.Element(XliffConstants.source);
            XElement targetElement = segmentElement.Element(XliffConstants.target);

            // segment pair properties
            ISegmentPairProperties props = ItemFactory.CreateSegmentPairProperties();

            // state -> confirmation level
            props.ConfirmationLevel = MapSegmentStateToConfirmationLevel(segmentElement.Attribute("state"));

            // id
            string id = (string)segmentElement.Attribute("id");
            if (id != null)
            {
                props.Id = new SegmentId(id);
            }

            // source content
            ISegment sourceSegment = ItemFactory.CreateSegment(props);
            paragraphUnit.Source.Add(sourceSegment);
            _bcmGeneratorRepository.GenerateContent(unitInfo, sourceElement.Nodes(), sourceSegment);
            
            // target content
            ISegment targetSegment = ItemFactory.CreateSegment(props);
            paragraphUnit.Target.Add(targetSegment);
            if (targetElement != null)
            {
                _bcmGeneratorRepository.GenerateContent(unitInfo, targetElement.Nodes(), targetSegment);
            }

            // segment comments
            XElement notesElement = segmentElement.Element(XliffConstants.notes);
            if (notesElement != null)
            {
                AddSegmentComments(notesElement, sourceSegment, targetSegment);
            }
        }

        #endregion <unit> element

        #region <group> element
        
        private void StartGroup()
        {
            // enter context
            ContextData contextData =
                XliffContext.CreateContextInfo(
                PropertiesFactory,
                _contextStack,
                _reader.GetAttribute("fs", XliffConstants.Xliff20FormatStyleNamespace), 
                true);
            _contextStack.Push(contextData);
        }

        private void EndGroup()
        {
            // exit context
            _contextStack.Pop();
        }

        private void SetContextProperties(IParagraphUnit paragraphUnit, XElement unitElement)
        {
            // add paragraph unit context based on fs attribute value
            XAttribute fsAttribute = unitElement.Attribute(XliffConstants.fs);

            ContextData puContextData = XliffContext.CreateContextInfo(PropertiesFactory, _contextStack, (string)fsAttribute, false);

            IContextProperties contextProperties = PropertiesFactory.CreateContextProperties();
            contextProperties.Contexts.Add(puContextData.Context);
            foreach (ContextData cd in _contextStack)
            {
                contextProperties.Contexts.Add(cd.Context);
            }

            // assign document structure context
            if (puContextData.StructureInfo != null)
            {
                contextProperties.StructureInfo = puContextData.StructureInfo;
            }
            else
            {
                contextProperties.StructureInfo = _contextStack.Select(cd => cd.StructureInfo).FirstOrDefault(si => si != null);
            }

            paragraphUnit.Properties.Contexts = contextProperties;
        }
        
        #endregion

        #region Helpers

        

        private List<OriginalData> ParseOriginalData(XElement originalDataElement)
        {
            if (originalDataElement == null)
            {
                return null;
            }
            // TODO: add support for <cp>
            return (from e in originalDataElement.Elements(XliffConstants.data)
                    select new OriginalData { Id = e.Attribute("id").Value, Contents = e.Value }).ToList();
        }

        private ConfirmationLevel MapSegmentStateToConfirmationLevel(XAttribute state)
        {
            if (state == null)
            {
                return ConfirmationLevel.Unspecified;
            }
            switch (state.Value)
            {
                case "initial":
                    return ConfirmationLevel.Unspecified;
                case "translated":
                    return ConfirmationLevel.Translated;
                case "reviewed":
                    return ConfirmationLevel.ApprovedTranslation;
                case "final":
                    return ConfirmationLevel.ApprovedSignOff;
                default:
                    throw new ArgumentException("Unrecognized state value: " + state.Value);
            }
        }

        private void CreateContentOutsideSegments(UnitInfo unitInfo, XElement ignorableElement, IParagraphUnit paragraphUnit)
        {

        }

        private ICommentProperties CreateUnitComments(XElement notesElement)
        {
            ICommentProperties comments = PropertiesFactory.CreateCommentProperties();

            foreach (XElement noteElement in notesElement.Elements(XliffConstants.note))
            {
                IComment comment = CreateComment(noteElement);
                comments.Add(comment);
            }
            return comments;
        }

        private void AddSegmentComments(XElement notesElement, ISegment sourceSegment, ISegment targetSegment)
        {
            ICommentMarker sourceCommentMarker = ItemFactory.CreateCommentMarker(PropertiesFactory.CreateCommentProperties());
            ICommentMarker targetCommentMarker = ItemFactory.CreateCommentMarker(PropertiesFactory.CreateCommentProperties());

            foreach (XElement noteElement in notesElement.Elements(XliffConstants.note))
            {
                // create comment
                IComment comment = CreateComment(noteElement);
                
                // source or target
                ICommentMarker marker;
                string appliesTo = (string)noteElement.Attribute("appliesTo");
                marker = (appliesTo == "source") ? sourceCommentMarker : targetCommentMarker;

                // add comment
                marker.Comments.Add(comment);
            }

            if (sourceCommentMarker != null)
            {
                AddSegmentCommentMarker(sourceSegment, sourceCommentMarker);
            }
            if (targetCommentMarker != null)
            {
                AddSegmentCommentMarker(targetSegment, targetCommentMarker);
            }
        }

        private IComment CreateComment(XElement noteElement)
        {
            // severity
            string priorityValue = (string)noteElement.Attribute("priority");
            int priority = !String.IsNullOrEmpty(priorityValue) ? Int32.Parse(priorityValue) : 5;
            Severity severity;
            if (priority == 1)
            {
                severity = Severity.High;
            }
            else if (priority <= 5)
            {
                severity = Severity.Medium;
            }
            else
            {
                severity = Severity.Low;
            }
            
            IComment comment = PropertiesFactory.CreateComment(noteElement.Value, "XLIFF 2.0", severity);
            comment.Version = "1.0";


            return comment;
        }

        private void AddSegmentCommentMarker(ISegment segment, ICommentMarker marker)
        {
            List<IAbstractMarkupData> segmentContents = segment.ToList();
            segment.MoveAllItemsTo(marker);
            segment.Add(marker);
        }

        #endregion Helpers

        #region Initialization
        public event EventHandler<ProgressEventArgs> Progress;
        public IDocumentProperties DocumentProperties { get; set; }
        public IBilingualContentHandler Output { get; set; }
        
        public void SetFileProperties(IFileProperties properties)
        {
            _fileProperties = properties;
            Output.Initialize(DocumentProperties);
            Output.SetFileProperties(properties);
        }

        
        public void EndOfInput()
        {
            //close document here
            _reader.Close();
            Output.FileComplete();
        }

        public void InitializeSettings(ISettingsBundle settingsBundle, string configurationId)
        {
            //_userSettings = new Settings();
            //_userSettings.PopulateFromSettingsBundle(settingsBundle, configurationId);
        }

        public void Dispose()
        {
        }

        #endregion    

    }
}
